@@ -2,14 +2,10 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) The Exim Maintainers 2020 - 2025 */
 /* Copyright (c) University of Cambridge 1995 - 2018 */
 /* See the file NOTICE for conditions of use and distribution. */
-/* SPDX-License-Identifier: GPL-2.0-or-later */
 
 #include "../exim.h"
-
-#ifdef ROUTER_QUERYPROGRAM	/* Remainder of file */
 #include "rf_functions.h"
 #include "queryprogram.h"
 
@@ -19,23 +15,23 @@
 
 optionlist queryprogram_router_options[] = {
   { "*expand_command_group", opt_bool | opt_hidden,
-      OPT_OFF(queryprogram_router_options_block, expand_cmd_gid) },
+      (void *)(offsetof(queryprogram_router_options_block, expand_cmd_gid)) },
   { "*expand_command_user", opt_bool | opt_hidden,
-      OPT_OFF(queryprogram_router_options_block, expand_cmd_uid) },
+      (void *)(offsetof(queryprogram_router_options_block, expand_cmd_uid)) },
   { "*set_command_group",   opt_bool | opt_hidden,
-      OPT_OFF(queryprogram_router_options_block, cmd_gid_set) },
+      (void *)(offsetof(queryprogram_router_options_block, cmd_gid_set)) },
   { "*set_command_user",    opt_bool | opt_hidden,
-      OPT_OFF(queryprogram_router_options_block, cmd_uid_set) },
+      (void *)(offsetof(queryprogram_router_options_block, cmd_uid_set)) },
   { "command",      opt_stringptr,
-      OPT_OFF(queryprogram_router_options_block, command) },
+      (void *)(offsetof(queryprogram_router_options_block, command)) },
   { "command_group",opt_expand_gid,
-      OPT_OFF(queryprogram_router_options_block, cmd_gid) },
+      (void *)(offsetof(queryprogram_router_options_block, cmd_gid)) },
   { "command_user", opt_expand_uid,
-      OPT_OFF(queryprogram_router_options_block, cmd_uid) },
+      (void *)(offsetof(queryprogram_router_options_block, cmd_uid)) },
   { "current_directory", opt_stringptr,
-      OPT_OFF(queryprogram_router_options_block, current_directory) },
+      (void *)(offsetof(queryprogram_router_options_block, current_directory)) },
   { "timeout",      opt_time,
-      OPT_OFF(queryprogram_router_options_block, timeout) }
+      (void *)(offsetof(queryprogram_router_options_block, timeout)) }
 };
 
 /* Size of the options list. An extern variable has to be used so that its
@@ -49,7 +45,7 @@ int queryprogram_router_options_count =
 
 /* Dummy entries */
 queryprogram_router_options_block queryprogram_router_option_defaults = {0};
-void queryprogram_router_init(driver_instance *rblock) {}
+void queryprogram_router_init(router_instance *rblock) {}
 int queryprogram_router_entry(router_instance *rblock, address_item *addr,
   struct passwd *pw, int verify, address_item **addr_local,
   address_item **addr_remote, address_item **addr_new,
@@ -82,21 +78,21 @@ queryprogram_router_options_block queryprogram_router_option_defaults = {
 consistency checks to be done, or anything else that needs to be set up. */
 
 void
-queryprogram_router_init(driver_instance * rblock)
+queryprogram_router_init(router_instance *rblock)
 {
 queryprogram_router_options_block *ob =
   (queryprogram_router_options_block *)(rblock->options_block);
 
 /* A command must be given */
 
-if (!ob->command)
-  log_write_die(0, LOG_CONFIG_FOR, "%s router:\n  "
+if (ob->command == NULL)
+  log_write(0, LOG_PANIC_DIE|LOG_CONFIG_FOR, "%s router:\n  "
     "a command specification is required", rblock->name);
 
 /* A uid/gid must be supplied */
 
-if (!ob->cmd_uid_set && !ob->expand_cmd_uid)
-  log_write_die(0, LOG_CONFIG_FOR, "%s router:\n  "
+if (!ob->cmd_uid_set && ob->expand_cmd_uid == NULL)
+  log_write(0, LOG_PANIC_DIE|LOG_CONFIG_FOR, "%s router:\n  "
     "command_user must be specified", rblock->name);
 }
 
@@ -121,9 +117,9 @@ Returns:         nothing
 */
 
 static void
-add_generated(router_instance * rblock, address_item ** addr_new,
-  address_item * addr, address_item * generated,
-  const address_item_propagated * addr_prop)
+add_generated(router_instance *rblock, address_item **addr_new,
+  address_item *addr, address_item *generated,
+  address_item_propagated *addr_prop)
 {
 while (generated != NULL)
   {
@@ -141,12 +137,12 @@ while (generated != NULL)
   *addr_new = next;
 
   if (addr->child_count == USHRT_MAX)
-    log_write_die(0, LOG_MAIN, "%s router generated more than %d "
-      "child addresses for <%s>", rblock->drinst.name, USHRT_MAX, addr->address);
+    log_write(0, LOG_MAIN|LOG_PANIC_DIE, "%s router generated more than %d "
+      "child addresses for <%s>", rblock->name, USHRT_MAX, addr->address);
   addr->child_count++;
 
   DEBUG(D_route)
-    debug_printf("%s router generated %s\n", rblock->drinst.name, next->address);
+    debug_printf("%s router generated %s\n", rblock->name, next->address);
   }
 }
 
@@ -215,8 +211,8 @@ uschar buffer[1024];
 const uschar **argvptr;
 uschar *rword, *rdata, *s;
 address_item_propagated addr_prop;
-queryprogram_router_options_block * ob =
-  (queryprogram_router_options_block *)(rblock->drinst.options_block);
+queryprogram_router_options_block *ob =
+  (queryprogram_router_options_block *)(rblock->options_block);
 uschar *current_directory = ob->current_directory;
 ugid_block ugid;
 uid_t curr_uid = getuid();
@@ -226,8 +222,8 @@ gid_t gid = ob->cmd_gid;
 uid_t *puid = &uid;
 gid_t *pgid = &gid;
 
-DEBUG(D_route) debug_printf_indent("%s router called for %s: domain = %s\n",
-  rblock->drinst.name, addr->address, addr->domain);
+DEBUG(D_route) debug_printf("%s router called for %s: domain = %s\n",
+  rblock->name, addr->address, addr->domain);
 
 ugid.uid_set = ugid.gid_set = FALSE;
 
@@ -236,7 +232,6 @@ errors address and extra header stuff. */
 
 bzero(&addr_prop, sizeof(addr_prop));
 addr_prop.address_data = deliver_address_data;
-tree_dup((tree_node **)&addr_prop.variables, addr->prop.variables);
 
 rc = rf_get_errors_address(addr, rblock, verify, &addr_prop.errors_address);
 if (rc != OK) return rc;
@@ -245,31 +240,41 @@ rc = rf_get_munge_headers(addr, rblock, &addr_prop.extra_headers,
   &addr_prop.remove_headers);
 if (rc != OK) return rc;
 
+#ifdef EXPERIMENTAL_SRS
+addr_prop.srs_sender = NULL;
+#endif
+
 /* Get the fixed or expanded uid under which the command is to run
 (initialization ensures that one or the other is set). */
 
-if (  !ob->cmd_uid_set
-   && !route_find_expanded_user(ob->expand_cmd_uid, rblock->drinst.name,
-	US"router", &upw, &uid, &(addr->message)))
+if (!ob->cmd_uid_set)
+  {
+  if (!route_find_expanded_user(ob->expand_cmd_uid, rblock->name, US"router",
+      &upw, &uid, &(addr->message)))
     return DEFER;
+  }
 
 /* Get the fixed or expanded gid, or take the gid from the passwd entry. */
 
 if (!ob->cmd_gid_set)
-  if (ob->expand_cmd_gid)
+  {
+  if (ob->expand_cmd_gid != NULL)
     {
-    if (route_find_expanded_group(ob->expand_cmd_gid, rblock->drinst.name,
+    if (route_find_expanded_group(ob->expand_cmd_gid, rblock->name,
         US"router", &gid, &(addr->message)))
       return DEFER;
     }
-  else if (upw)
+  else if (upw != NULL)
+    {
     gid = upw->pw_gid;
+    }
   else
     {
     addr->message = string_sprintf("command_user set without command_group "
-      "for %s router", rblock->drinst.name);
+      "for %s router", rblock->name);
     return DEFER;
     }
+  }
 
 DEBUG(D_route) debug_printf("requires uid=%ld gid=%ld current_directory=%s\n",
   (long int)uid, (long int)gid, current_directory);
@@ -289,23 +294,26 @@ if (curr_uid != root_uid && (uid != curr_uid || gid != curr_gid))
 
 /* Set up the command to run */
 
-GET_OPTION("command");
 if (!transport_set_up_command(&argvptr, /* anchor for arg list */
     ob->command,                        /* raw command */
-    TSUC_EXPAND_ARGS,                   /* arguments expanded but must not be tainted */
+    TRUE,                               /* expand the arguments */
     0,                                  /* not relevant when... */
     NULL,                               /* no transporting address */
     US"queryprogram router",            /* for error messages */
-    &addr->message))                    /* where to put error message */
+    &(addr->message)))                  /* where to put error message */
+  {
   return DEFER;
+  }
 
 /* Create the child process, making it a group leader. */
 
-if ((pid = child_open_uid(argvptr, NULL, 0077, puid, pgid, &fd_in, &fd_out,
-			  current_directory, TRUE, US"queryprogram-cmd")) < 0)
+pid = child_open_uid(argvptr, NULL, 0077, puid, pgid, &fd_in, &fd_out,
+  current_directory, TRUE);
+
+if (pid < 0)
   {
   addr->message = string_sprintf("%s router couldn't create child process: %s",
-    rblock->drinst.name, strerror(errno));
+    rblock->name, strerror(errno));
   return DEFER;
   }
 
@@ -320,22 +328,22 @@ if ((rc = child_close(pid, ob->timeout)) != 0)
   {
   if (rc > 0)
     addr->message = string_sprintf("%s router: command returned non-zero "
-      "code %d", rblock->drinst.name, rc);
+      "code %d", rblock->name, rc);
 
   else if (rc == -256)
     {
     addr->message = string_sprintf("%s router: command timed out",
-      rblock->drinst.name);
+      rblock->name);
     killpg(pid, SIGKILL);       /* Kill the whole process group */
     }
 
   else if (rc == -257)
     addr->message = string_sprintf("%s router: wait() failed: %s",
-      rblock->drinst.name, strerror(errno));
+      rblock->name, strerror(errno));
 
   else
     addr->message = string_sprintf("%s router: command killed by signal %d",
-      rblock->drinst.name, -rc);
+      rblock->name, -rc);
 
   return DEFER;
   }
@@ -350,7 +358,7 @@ len = read(fd_out, buffer, sizeof(buffer) - 1);
 if (len <= 0)
   {
   addr->message = string_sprintf("%s router: command failed to return data",
-    rblock->drinst.name);
+    rblock->name);
   return DEFER;
   }
 
@@ -363,10 +371,10 @@ buffer[len] = 0;
 DEBUG(D_route) debug_printf("command wrote: %s\n", buffer);
 
 rword = buffer;
-Uskip_whitespace(&rword);
+while (isspace(*rword)) rword++;
 rdata = rword;
-Uskip_nonwhite(&rdata);
-if (*rdata) *rdata++ = '\0';
+while (*rdata != 0 && !isspace(*rdata)) rdata++;
+if (*rdata != 0) *rdata++ = 0;
 
 /* The word must be a known yield name. If it is "REDIRECT", the rest of the
 line is redirection data, as for a .forward file. It may not contain filter
@@ -388,13 +396,16 @@ if (strcmpic(rword, US"REDIRECT") == 0)
       RDO_INCLUDE |              /* forbid :include: */
       RDO_REWRITE,               /* rewrite generated addresses */
     NULL,                        /* :include: directory not relevant */
-    NULL,                        /* sieve info not relevant */
+    NULL,                        /* sieve vacation directory not relevant */
+    NULL,                        /* sieve enotify mailto owner not relevant */
+    NULL,                        /* sieve useraddress not relevant */
+    NULL,                        /* sieve subaddress not relevant */
     &ugid,                       /* uid/gid (but not set) */
     &generated,                  /* where to hang the results */
-    &addr->message,              /* where to put messages */
+    &(addr->message),            /* where to put messages */
     NULL,                        /* don't skip syntax errors */
     &filtertype,                 /* not used; will always be FILTER_FORWARD */
-    string_sprintf("%s router", rblock->drinst.name));
+    string_sprintf("%s router", rblock->name));
 
   switch (rc)
     {
@@ -403,28 +414,28 @@ if (strcmpic(rword, US"REDIRECT") == 0)
     response after verifying. */
 
     case FF_DEFER:
-      if (!addr->message) addr->message = US"forced defer";
+    if (addr->message == NULL) addr->message = US"forced defer";
       else addr->user_message = addr->message;
-      return DEFER;
+    return DEFER;
 
     case FF_FAIL:
-      add_generated(rblock, addr_new, addr, generated, &addr_prop);
-      if (!addr->message) addr->message = US"forced rejection";
+    add_generated(rblock, addr_new, addr, generated, &addr_prop);
+    if (addr->message == NULL) addr->message = US"forced rejection";
       else addr->user_message = addr->message;
-      return FAIL;
+    return FAIL;
 
     case FF_DELIVERED:
-      break;
+    break;
 
     case FF_NOTDELIVERED:    /* an empty redirection list is bad */
-      addr->message = US"no addresses supplied";
+    addr->message = US"no addresses supplied";
     /* Fall through */
 
     case FF_ERROR:
     default:
-      addr->basic_errno = ERRNO_BADREDIRECT;
-      addr->message = string_sprintf("error in redirect data: %s", addr->message);
-      return DEFER;
+    addr->basic_errno = ERRNO_BADREDIRECT;
+    addr->message = string_sprintf("error in redirect data: %s", addr->message);
+    return DEFER;
     }
 
   /* Handle the generated addresses, if any. */
@@ -456,30 +467,35 @@ if (strcmpic(rword, US"accept") != 0)
   else if (strcmpic(rword, US"defer") != 0)
     {
     addr->message = string_sprintf("bad command yield: %s %s", rword, rdata);
-    log_write(0, LOG_PANIC, "%s router: %s", rblock->drinst.name, addr->message);
+    log_write(0, LOG_PANIC, "%s router: %s", rblock->name, addr->message);
     }
   return DEFER;
   }
 
 /* The command yielded "ACCEPT". The rest of the string is a number of keyed
-fields from which we can fish out values using the equivalent of the "extract"
-expansion function. */
+fields from which we can fish out values using the "extract" expansion
+function. To use this feature, we must put the string into the $value variable,
+i.e. set lookup_value. */
 
-if ((s = expand_getkeyed(US"data", rdata)) && *s)
-  addr_prop.address_data = string_copy(s);
+lookup_value = rdata;
+s = expand_string(US"${extract{data}{$value}}");
+if (*s != 0) addr_prop.address_data = string_copy(s);
+
+s = expand_string(US"${extract{transport}{$value}}");
+lookup_value = NULL;
 
 /* If we found a transport name, find the actual transport */
 
-if ((s = expand_getkeyed(US"transport", rdata)) && *s)
+if (*s != 0)
   {
-  transport_instance * transport;
-  for (transport = transports; transport; transport = transport->drinst.next)
-    if (Ustrcmp(transport->drinst.name, s) == 0) break;
-  if (!transport)
+  transport_instance *transport;
+  for (transport = transports; transport != NULL; transport = transport->next)
+    if (Ustrcmp(transport->name, s) == 0) break;
+  if (transport == NULL)
     {
     addr->message = string_sprintf("unknown transport name %s yielded by "
       "command", s);
-    log_write(0, LOG_PANIC, "%s router: %s", rblock->drinst.name, addr->message);
+    log_write(0, LOG_PANIC, "%s router: %s", rblock->name, addr->message);
     return DEFER;
     }
   addr->transport = transport;
@@ -491,27 +507,32 @@ the last argument not being NULL. */
 
 else
   {
-  if (!rf_get_transport(rblock->transport_name, &rblock->transport, addr,
-       rblock->drinst.name, US"transport"))
+  if (!rf_get_transport(rblock->transport_name, &(rblock->transport), addr,
+       rblock->name, US"transport"))
     return DEFER;
   addr->transport = rblock->transport;
   }
 
 /* See if a host list is given, and if so, look up the addresses. */
 
-if ((s = expand_getkeyed(US"hosts", rdata)) && *s)
+lookup_value = rdata;
+s = expand_string(US"${extract{hosts}{$value}}");
+
+if (*s != 0)
   {
   int lookup_type = LK_DEFAULT;
-  uschar * ss = expand_getkeyed(US"lookup", rdata);
+  uschar *ss = expand_string(US"${extract{lookup}{$value}}");
+  lookup_value = NULL;
 
-  if (ss && *ss)
+  if (*ss != 0)
     {
     if (Ustrcmp(ss, "byname") == 0) lookup_type = LK_BYNAME;
     else if (Ustrcmp(ss, "bydns") == 0) lookup_type = LK_BYDNS;
     else
       {
-      addr->message = string_sprintf("bad lookup type %q yielded by command", ss);
-      log_write(0, LOG_PANIC, "%s router: %s", rblock->drinst.name, addr->message);
+      addr->message = string_sprintf("bad lookup type \"%s\" yielded by "
+        "command", ss);
+      log_write(0, LOG_PANIC, "%s router: %s", rblock->name, addr->message);
       return DEFER;
       }
     }
@@ -530,33 +551,9 @@ addr->prop = addr_prop;
 
 /* Queue the address for local or remote delivery. */
 
-return rf_queue_add(addr, addr_local, addr_remote, rblock, pw) ? OK : DEFER;
+return rf_queue_add(addr, addr_local, addr_remote, rblock, pw)?
+  OK : DEFER;
 }
 
-
-
-# ifdef DYNLOOKUP
-#  define queryprogram_router_info _router_info
-# endif
-#endif	/*!MACRO_PREDEF*/
-
-router_info queryprogram_router_info =
-{
-.drinfo = {
-  .driver_name =      US"queryprogram",
-  .options =          queryprogram_router_options,
-  .options_count =    &queryprogram_router_options_count,
-  .options_block =    &queryprogram_router_option_defaults,
-  .options_len =      sizeof(queryprogram_router_options_block),
-  .init =             queryprogram_router_init,
-# ifdef DYNLOOKUP
-  .dyn_magic =	      ROUTER_MAGIC,
-# endif
-  },
-.code =               queryprogram_router_entry,
-.tidyup =             NULL,     /* no tidyup entry */
-.ri_flags =           0
-};
-
-#endif	/*ROUTER_QUERYPROGRAM*/
+#endif   /*!MACRO_PREDEF*/
 /* End of routers/queryprogram.c */
